﻿package supermotion3d
{
	import flash.display.Sprite;
	import flash.display.Bitmap;
	import flash.display.BitmapData;
	import flash.display.TriangleCulling;
	import flash.geom.Matrix3D;
	import flash.geom.Vector3D;
	import flash.geom.Utils3D;

	public class HierarchyModel3D extends Model3D
	{
		//タイム
		private var _preTime:int = 0;

		public function HierarchyModel3D(sprite:Sprite = null):void
		{
			super(sprite);
		}

		override protected function init(texture:BitmapData):void
		{
			var b:int = 0;
			var i:int = 0;
			var j:int = 0;
			var index:int = 0;
			var indices:Vector.<int> = new Vector.<int>();

			//テクスチャ画像
			_image = texture;

			//ボーンデータ
			for ( i = 0; i < _boneNum; i++ )
			{
				_vertices3D.push(new Vector.<Number>());
				_uvtData.push(new Vector.<Number>());
				indices.push(0);
				_transform.push(new Matrix3D());
			}
			//ボーンごとに属する頂点とUV座標
			for ( i = 0; i < _indices.length; i++ )
			{
				b = _boneIndices[_indices[i]];
				_vertices3D[b].push(
					_vertices3D2[_indices[i]*3  ],
					_vertices3D2[_indices[i]*3+1],
					_vertices3D2[_indices[i]*3+2]);
				_uvtData[b].push(
					_uvtData2[i*3  ],
					_uvtData2[i*3+1],
					_uvtData2[i*3+2]);
			}
			//ボーンインデックス
			for ( i = 0; i < _indices.length; i++ )
			{
				var length:int = 0;
				b = _boneIndices[_indices[i]];
				for ( j = 0; j < b; j++ )
				{
					length += _vertices3D[j].length/3;
				}
				_indices2.push(indices[b]+length);
				indices[b] += 1;
			}
			_currentAnimation = 0;
		}

		protected function addVertexBone(x:Number,y:Number,z:Number,i:int):void
		{
			_vertices3D2.push(x,y,z);
			_boneIndices.push(i);
		}

		override protected function addUV(u:Number,v:Number):void
		{
			_uvtData2.push(u,v,0);
		}

		protected function addIndex(i0:int,i1:int,i2:int):void
		{
			_indices.push(i0,i1,i2);
		}

		protected function addPivot(p:int,pX:Number,pY:Number,pZ:Number):void
		{
			var current:int;
			_parents.push(p);
			_pivots.push(new Vector.<Number>());
			current = _pivots.length - 1;
			_pivots[current].push(pX, pY, pZ);
			_boneNum = _pivots.length;
		}

		protected function addAnimation3():void
		{
			_translate.push(new Vector.<Vector.<Vector3D>>());
			_rotate.push(new Vector.<Vector.<Vector3D>>());
			_scaling.push(new Vector.<Vector.<Vector3D>>());
			_currentAnimation = _translate.length - 1;
		}

		protected function addAnimation2():void
		{
			_translate[_currentAnimation].push(new Vector.<Vector3D>());
			_rotate[_currentAnimation].push(new Vector.<Vector3D>());
			_scaling[_currentAnimation].push(new Vector.<Vector3D>());
			_current = _translate[_currentAnimation].length - 1;
			_preTime = 0;
		}

		protected function addAnimation(
			index:int,
			tX:Number,tY:Number,tZ:Number,
			rX:Number,rY:Number,rZ:Number,
			sX:Number,sY:Number,sZ:Number):void
		{
			var k:int;
			var translate:Vector3D = new Vector3D(tX,tY,tZ);
			var rotation:Vector3D = new Vector3D(rX,rY,rZ);
			if ( sX <= 0 )
			{
				sX = 0.000001;
			}
			if ( sY <= 0 )
			{
				sY = 0.000001;
			}
			if ( sZ <= 0 )
			{
				sZ = 0.000001;
			}
			var scaling:Vector3D = new Vector3D(sX,sY,sZ);

			//0フレーム目の場合
			if ( index == 0 )
			{
				_translate[_currentAnimation][_current].push(translate);
				_rotate[_currentAnimation][_current].push(rotation);
				_scaling[_currentAnimation][_current].push(scaling);
				return;
			}

			//1フレーム目以降の場合
			var length:int = index - _preTime;

			//平行移動アニメーションの補間
			var trans:Vector3D = new Vector3D();
			var preTrans:Vector3D =
				_translate[_currentAnimation][_current][_preTime];
			trans.x = translate.x - preTrans.x;
			trans.y = translate.y - preTrans.y;
			trans.z = translate.z - preTrans.z;
			for ( k = 1; k <= length; k++ )
			{
				var trans2:Vector3D = new Vector3D();
				trans2.x=preTrans.x+Number(Number(k)/length)*trans.x;
				trans2.y=preTrans.y+Number(Number(k)/length)*trans.y;
				trans2.z=preTrans.z+Number(Number(k)/length)*trans.z;
				_translate[_currentAnimation][_current].push(trans2);
			}

			//回転アニメーションの補間
			var rotate:Vector3D = new Vector3D();
			var preRotate:Vector3D =
				_rotate[_currentAnimation][_current][_preTime];
			rotate.x = rotation.x - preRotate.x;
			rotate.y = rotation.y - preRotate.y;
			rotate.z = rotation.z - preRotate.z;
			for ( k = 1; k <= length; k++ )
			{
				var rotate2:Vector3D = new Vector3D();
				rotate2.x=preRotate.x+Number(Number(k)/length)*rotate.x;
				rotate2.y=preRotate.y+Number(Number(k)/length)*rotate.y;
				rotate2.z=preRotate.z+Number(Number(k)/length)*rotate.z;
				_rotate[_currentAnimation][_current].push(rotate2);
			}

			//スケーリングアニメーションの補間
			var scale:Vector3D = new Vector3D();
			var preScale:Vector3D =
				_scaling[_currentAnimation][_current][_preTime];
			scale.x = scaling.x - preScale.x;
			scale.y = scaling.y - preScale.y;
			scale.z = scaling.z - preScale.z;
			for ( k = 1; k <= length; k++ )
			{
				var scale2:Vector3D = new Vector3D();
				scale2.x = preScale.x+Number(Number(k)/length)*scale.x;
				scale2.y = preScale.y+Number(Number(k)/length)*scale.y;
				scale2.z = preScale.z+Number(Number(k)/length)*scale.z;
				_scaling[_currentAnimation][_current].push(scale2);
			}

			_preTime = index;
		}

		override public function draw(camera:Matrix3D,projection:Matrix3D):void
		{
			var i:int;
			var indices:Vector.<int> = new Vector.<int>();
			var faces:Array = new Array();
			var faces2:Array = new Array();
			//行列
			var m:Matrix3D = camera.clone();
			var vertices3D:Vector.<Number> = new Vector.<Number>();
			var vertices2D:Vector.<Number> = new Vector.<Number>();
			var uv:Vector.<Number> = new Vector.<Number>();
			var uvtData:Vector.<Number> = Vector.<Number>([0, 0, 0, 1, 1, 0, 1, 1]); 

			//画像消去
			this.graphics.clear();

			//座標変換
			m.prepend(_matrix);
			//トランスフォーム行列設定
			setTransform(-1,m);
			
			//トランスフォーム
			for ( i = 0; i < _boneNum; i++ )
			{
				//内部で使用する頂点
				var vert3D:Vector.<Number> = new Vector.<Number>();
				//2D変換後の頂点
				var vert2D:Vector.<Number> = new Vector.<Number>();

				//頂点のトランスフォーム
				_transform[i].transformVectors(_vertices3D[i], vert3D);
				Utils3D.projectVectors(projection, vert3D, vert2D, _uvtData[i]);
				vertices2D = vertices2D.concat(vert2D);
				uv = uv.concat(_uvtData[i]);
				vertices3D = vertices3D.concat(vert3D);
			}
			//Z値を追加
			for ( i = 0; i < _indices2.length; i += 3 )
			{
				faces.push(
					(vertices3D[_indices2[i  ]*3+2]+
					 vertices3D[_indices2[i+1]*3+2]+
					 vertices3D[_indices2[i+2]*3+2])/3);
			}
			// Zソート (奥のものから順番に並べる)
			faces2 = faces.sort(Array.DESCENDING | Array.NUMERIC | Array.RETURNINDEXEDARRAY);
			for each( i in faces2 )
			{
				if ( faces[i] > 200 )
				{
					indices.push(_indices2[i*3],_indices2[i*3+1],_indices2[i*3+2]);
				}
			}

			//3D描画
			if (indices.length > 0)
			{
				//テクスチャ
				this.graphics.beginBitmapFill(_image);
				this.graphics.drawTriangles(
					vertices2D, indices, uv, TriangleCulling.POSITIVE);
				this.graphics.endFill();
			}
		}

		//トランスフォーム
		protected function setTransform(parent:int,m:Matrix3D):void
		{
			for ( var i:int = 0; i < _pivots.length; i++ )
			{
				if ( _parents[i] == parent )
				{
					_transform[i].identity();
					_transform[i].prepend(m);
					_transform[i].prependTranslation(
						_pivots[i][0]+_translate[_currentAnimation][i][_time].x,
						_pivots[i][1]+_translate[_currentAnimation][i][_time].y,
						_pivots[i][2]+_translate[_currentAnimation][i][_time].z);
					_transform[i].prependScale(
						_scaling[_currentAnimation][i][_time].x,
						_scaling[_currentAnimation][i][_time].y,
						_scaling[_currentAnimation][i][_time].z);
					_transform[i].prependRotation(
						_rotate[_currentAnimation][i][_time].x, Vector3D.X_AXIS);
					_transform[i].prependRotation(
						_rotate[_currentAnimation][i][_time].y, Vector3D.Y_AXIS);
					_transform[i].prependRotation(
						_rotate[_currentAnimation][i][_time].z, Vector3D.Z_AXIS);
					_transform[i].prependTranslation(
						-_pivots[i][0],-_pivots[i][1],-_pivots[i][2]);
					//再帰
					setTransform(i,_transform[i]);
				}
			}
		}

		//アニメーション再生
		override public function playAnimation():void
		{
			if ( !_stopAnimation )
			{
				_time++;
			}
			if ( _translate.length > _currentAnimation &&
				_translate[_currentAnimation].length > 0 &&
				_time >= _translate[_currentAnimation][0].length )
			{
				if ( _animationLast )
				{
					_time = _translate[_currentAnimation][0].length - 1;
					_stopAnimation = true;
				}
				else 
				{
					_time = 0;	
					if ( !_loopAnimation )
					{
						_stopAnimation = true;
					}
				}
			}
		}

		override public function set currentAnimation( value:int ):void
		{
			this._time = 0;
			if ( _translate.length > value )
			{
				this._currentAnimation = value;
			}
			else
			{
				this._currentAnimation = 0;
			}
		}
	}
}
